/*
 * Risintech serial touchscreen driver
 *
 * Copyright (c) 2013 Vicki Chen
 */


#include <linux/errno.h>
#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/slab.h>
#include <linux/input.h>
#include <linux/serio.h>
#include <linux/init.h>

#define DRIVER_DESC	"Risin serial touchscreen driver"

MODULE_AUTHOR("Risintech");
MODULE_DESCRIPTION(DRIVER_DESC);
MODULE_LICENSE("GPL");

/*
 * Definitions & global arrays.
 */

#define RTOUCH_FORMAT_TABLET_STATUS_BIT 0x80
#define RTOUCH_FORMAT_TABLET_TOUCH_BIT 0x40
#define RTOUCH_FORMAT_TABLET_LENGTH 5
#define RTOUCH_RESPONSE_BEGIN_BYTE 0x01
#define RTOUCH_RESPONSE_END_BYTE 0x0d

/* todo: check specs for max length of all responses */
#define RTOUCH_MAX_LENGTH 8

#define RTOUCH_MIN_XC 0
#define RTOUCH_MAX_XC 0xfff
#define RTOUCH_MIN_YC 0
#define RTOUCH_MAX_YC 0xfff

#define RTOUCH_GET_XC(data) (((data[2])<<7) + data[1])
#define RTOUCH_GET_YC(data) (((data[4])<<7) + data[3])
#define RTOUCH_GET_TOUCHED(data) (RTOUCH_FORMAT_TABLET_TOUCH_BIT & data[0])

/*
 * Per-touchscreen data.
 */

struct rtouch {
	struct input_dev *dev;
	struct serio *serio;
	int idx;
	unsigned char data[RTOUCH_MAX_LENGTH];
	char phys[32];
};

char RightFlag=0;

static void rtouch_process_format_tablet(struct rtouch *rtouch)
{
	struct input_dev *dev = rtouch->dev;
	int X,Y;
	if (RTOUCH_FORMAT_TABLET_LENGTH == ++rtouch->idx) {

		X = (((int)rtouch->data[1]) | (((int)rtouch->data[2])<<7));//<<4;
		Y = (((int)rtouch->data[3]) | (((int)rtouch->data[4])<<7));//<<4;
		//printk(KERN_INFO "X=%d Y=%d\n",X,Y);

		switch (rtouch->data[0])
		{
		case 0x98:	
			if (RightFlag==0)
			{
				input_report_key(dev, BTN_TOUCH, 0);
				printk(KERN_INFO "left up\n");
			}
			else
			{
				RightFlag=0;	
				input_report_key(dev, BTN_STYLUS2, 0);
				printk(KERN_INFO "right up\n");
			}
			break;
		case 0xd8:
			input_report_key(dev, BTN_TOUCH, 1);
			printk(KERN_INFO "left down\n");
			break;
		case 0x99:
			
			if (RightFlag==0)
			{
				RightFlag=1;
				input_report_key(dev, BTN_STYLUS2, 1);
				printk(KERN_INFO "right down\n");
			}		
			break;
		}
						
		input_report_abs(dev, ABS_X, X);
		input_report_abs(dev, ABS_Y, Y);
		
		input_sync(dev);
		rtouch->idx = 0;
	}
}

static void rtouch_process_response(struct rtouch *rtouch)
{
	if (RTOUCH_RESPONSE_END_BYTE == rtouch->data[rtouch->idx++]) {
		/* FIXME - process response */
		rtouch->idx = 0;
	} else if (RTOUCH_MAX_LENGTH == rtouch->idx) {
		printk(KERN_ERR "rtouch.c: too many response bytes\n");
		rtouch->idx = 0;
	}
}

static irqreturn_t rtouch_interrupt(struct serio *serio,
		unsigned char data, unsigned int flags)
{
	struct rtouch* rtouch = serio_get_drvdata(serio);

	rtouch->data[rtouch->idx] = data;
	
	printk(KERN_INFO "[0]=%x [1]=%x [2]=%x [3]=%x [4]=%x [5]=%x [6]=%x [7]=%x\n",
			rtouch->data[0],rtouch->data[1],rtouch->data[2],rtouch->data[3],rtouch->data[4],
			rtouch->data[5],rtouch->data[6],rtouch->data[7]);
	
	
	if (RTOUCH_FORMAT_TABLET_STATUS_BIT & rtouch->data[0])
		rtouch_process_format_tablet(rtouch);
	else if (RTOUCH_RESPONSE_BEGIN_BYTE == rtouch->data[0])
		rtouch_process_response(rtouch);
	else
		printk(KERN_INFO "rtouch.c: unknown/unsynchronized data from device, byte %x\n",rtouch->data[0]);

	return IRQ_HANDLED;
}

/*
 * rtouch_disconnect() is the opposite of rtouch_connect()
 */

static void rtouch_disconnect(struct serio *serio)
{
	struct rtouch* rtouch = serio_get_drvdata(serio);

	input_get_device(rtouch->dev);
	input_unregister_device(rtouch->dev);
	serio_close(serio);
	serio_set_drvdata(serio, NULL);
	input_put_device(rtouch->dev);
	kfree(rtouch);
}

/*
 * rtouch_connect() is the routine that is called when someone adds a
 * new serio device that supports MicroTouch (Format Tablet) protocol and registers it as
 * an input device.
 */

static int rtouch_connect(struct serio *serio, struct serio_driver *drv)
{
	struct rtouch *rtouch;
	struct input_dev *input_dev;
	int err;
	printk(KERN_INFO "[rtouch]rtouch_connect\n");
	rtouch = kzalloc(sizeof(struct rtouch), GFP_KERNEL);
	input_dev = input_allocate_device();
	if (!rtouch || !input_dev) {
		err = -ENOMEM;
		goto fail1;
	}

	rtouch->serio = serio;
	rtouch->dev = input_dev;
	snprintf(rtouch->phys, sizeof(rtouch->phys), "%s/input0", serio->phys);

	input_dev->name = "Risintech Serial TouchScreen";
	input_dev->phys = rtouch->phys;
	input_dev->id.bustype = BUS_RS232;
	input_dev->id.vendor = SERIO_UNKNOWN;
	input_dev->id.product = 0x1ddb;
	input_dev->id.version = 0x0001;
	input_dev->dev.parent = &serio->dev;
	input_dev->evbit[0] = BIT_MASK(EV_KEY) | BIT_MASK(EV_ABS);
	input_dev->keybit[BIT_WORD(BTN_TOUCH)] = BIT_MASK(BTN_TOUCH) | BIT_MASK(BTN_STYLUS2);
	input_set_abs_params(rtouch->dev, ABS_X, RTOUCH_MIN_XC, RTOUCH_MAX_XC, 0, 0);
	input_set_abs_params(rtouch->dev, ABS_Y, RTOUCH_MIN_YC, RTOUCH_MAX_YC, 0, 0);

	serio_set_drvdata(serio, rtouch);

	err = serio_open(serio, drv);
	if (err)
	{
		printk(KERN_INFO "[rtouch]serio_open error!!\n");
		goto fail2;
	}

	err = input_register_device(rtouch->dev);
	if (err)
	{
		printk(KERN_INFO "[rtouch]input_register_device error!!\n");
		goto fail3;
	}
	printk(KERN_INFO "[rtouch]input_register_device pass!!\n");
	return 0;

 fail3:	serio_close(serio);
 fail2:	serio_set_drvdata(serio, NULL);
 fail1:	input_free_device(input_dev);
	kfree(rtouch);
	return err;
}

/*
 * The serio driver structure.
 */

static struct serio_device_id rtouch_serio_ids[] = {
	{
		.type	= SERIO_RS232,
		.proto	= SERIO_UNKNOWN,
		.id	= SERIO_ANY,
		.extra	= SERIO_ANY,
	},
	{ 0 }
};

MODULE_DEVICE_TABLE(serio, rtouch_serio_ids);

static struct serio_driver rtouch_drv = {
	.driver		= {
		.name	= "rtouch",
	},
	.description	= DRIVER_DESC,
	.id_table	= rtouch_serio_ids,
	.interrupt	= rtouch_interrupt,
	.connect	= rtouch_connect,
	.disconnect	= rtouch_disconnect,
};

/*
 * The functions for inserting/removing us as a module.
 */

static int __init rtouch_init(void)
{
	return serio_register_driver(&rtouch_drv);
}

static void __exit rtouch_exit(void)
{
	serio_unregister_driver(&rtouch_drv);
}

module_init(rtouch_init);
module_exit(rtouch_exit);
